/**
 * Sesame-Device
 * Christoph Hagen, 2022.
 *
 * The code for a simple door unlock mechanism where a servo pushes on an existing
 * physical button.
 */
#include <Arduino.h>
#include <WiFi.h>

#include "crypto.h"
#include "fresh.h"
#include "message.h"
#include "server.h"
#include "servo.h"
#include "config.h"

/* Global variables */

ServerConnection server(serverUrl, serverPort, serverPath);

ServoController servo(pwmTimer, servoFrequency, servoPin);

/* Event callbacks */

SesameEvent handleReceivedMessage(AuthenticatedMessage* payload, AuthenticatedMessage* response);

/* Logic */

void setup() {
    Serial.begin(serialBaudRate);
    Serial.setDebugOutput(true);
    Serial.println("[INFO] Device started");

    servo.configure(lockOpeningDuration, servoPressedState, servoReleasedState);
    Serial.println("[INFO] Servo configured");

    prepareMessageCounterUsage();
    //resetMessageCounter();
    printMessageCounter();

    server.onMessage(handleReceivedMessage);
}

uint32_t nextWifiReconnect = 0;
bool isReconnecting = false;

void loop() {
    uint32_t time = millis();

    server.loop();
    servo.loop();
    
    // Reconnect to WiFi
    if(time > nextWifiReconnect && WiFi.status() != WL_CONNECTED) {
        Serial.println("[INFO] Reconnecting WiFi...");
        WiFi.begin(wifiSSID, wifiPassword);
        isReconnecting = true;
        nextWifiReconnect = time + wifiReconnectInterval;
    }
    if (isReconnecting && WiFi.status() == WL_CONNECTED) {
        isReconnecting = false;
        Serial.println("[INFO] WiFi connected, opening socket");
        server.connectSSL(serverAccessKey);
        configureNTP(timeOffsetToGMT, timeOffsetDaylightSavings, ntpServerUrl);
        printLocalTime();
    }
}

SesameEvent processMessage(AuthenticatedMessage* message) {
    if (!isMessageCounterValid(message->message.id)) {
        return SesameEvent::MessageCounterInvalid;
    }
    if (!isMessageTimeAcceptable(message->message.time)) {
        return SesameEvent::MessageTimeMismatch;
    }
    if (!isAuthenticMessage(message, remoteKey, keySize)) {
        return SesameEvent::MessageAuthenticationFailed;
    }
    return SesameEvent::MessageAccepted;
}

/**
 * Process received binary data.
 *
 * Checks whether the received data is a valid and unused key,
 * and then signals that the motor should move.
 *
 * @param payload The pointer to the received data.
 * @param length The number of bytes received.
 * @return The event to signal to the server.
 */
SesameEvent handleReceivedMessage(AuthenticatedMessage* message, AuthenticatedMessage* response) {
    SesameEvent event = processMessage(message);
    
    // Only open when message is valid
    if (event == SesameEvent::MessageAccepted) {
        didUseMessageCounter(message->message.id);
        // Move servo
        servo.pressButton();
        Serial.printf("[Info] Accepted message %d\n", message->message.id);
    }

    // Create response for all cases
    response->message.time = getEpochTime();
    response->message.id = getNextMessageCounter();
    if (!authenticateMessage(response, localKey, keySize)) {
        return SesameEvent::MessageAuthenticationFailed;
    }
    return event;
}
